from abc import ABC 
from abc import abstractmethod
from openai import OpenAI
from typing import Any
import logging

class AiWrapper(ABC):
    """
    Abstract base class for AI wrappers.
    """

    @abstractmethod
    def call(self, input: str, **kwargs) -> str:
        """
        Abstract method to be implemented by subclasses. 

        Args:
            input (str): The input string.
            **kwargs: Optional keyword arguments.

        Returns:
            str: The output string.
        """
        raise NotImplementedError


class OpenAIWrapper(AiWrapper):
    """
    Wrapper for the OpenAI API.
    """

    def __init__(self, api_key: str, model: str = "gpt-3.5-turbo") -> None:
        """
        Initialize the OpenAIWrapper with the given API key and model.

        Args:
            api_key (str): The API key for OpenAI.
            model (str, optional): The model to use. Defaults to "gpt-3.5-turbo".
        """
        self.api_key = api_key
        self.model = model

    def call(self, input:str) -> str: # type: ignore
        """
        Call the OpenAI API with the given input and return the response.

        Args:
            input (str): The input string.

        Returns:
            str: The output string.
        """
        client = OpenAI(api_key=self.api_key)
        resp = client.chat.completions.create(
            model=self.model,
            messages=[{"role": "user", "content": input}],
            stream=False
        )
        response = resp.choices[0].message.content
        return str(response)

class SmartAiWrapper(AiWrapper):
    """
    Wrapper for complex AI operations.
    """

    def __init__(self, ai: AiWrapper, max_attempts: int = 3) -> None:
        """
        Initialize the ComplexAiWrapper with the given AI and maximum number of attempts.

        Args:
            ai (AiWrapper): The AI wrapper to use.
            max_attempts (int, optional): The maximum number of attempts. Defaults to 3.
        """
        self.ai = ai
        self.original_input: str = ""
        self._current_transformation_instructions: str = "1. Sort the data\n2. Perform the final transformation\n3. Return the data\n"
        self._previous_transformation_result: str = "No response"
        self.previous_transformation_error: str = "No Error"
        self.attempts: int = 0
        self.max_attempts: int = max_attempts
    
    def _come_up_with_transformation(self):
        """
        Come up with a new transformation based on the previous result.
        """
        prompt=(f"You are an AI who creates data transformation instructions for another AI based on the following DATA TRANSFORMATION REQUEST: {self.original_input}\n"
            f"You previously came up with these instructions:\n"
            f"{self._current_transformation_instructions}\n"
            f"The previous transformation result was:\n"
            f"{self._previous_transformation_result}\n"
            f"The previous transformation failed with the error:\n"
            f"{self.previous_transformation_error}\n"
            f"Please come up with new transformation instructions based on the previous result and error."
            )
        result = self.ai.call(prompt)
        logging.debug(result)
        
        self._current_transformation_instructions = self.ai.call(prompt)
    

    def _perform_current_transformation(self) -> Any:
        """
        Perform the current transformation and store the result.

        Returns:
            Any: The result of the transformation.
        """
        prompt = (
            f"You are an AI who performs data transformations based on the following instructions\n"
            f"{self._current_transformation_instructions}\n"
            f"Here is the input data:\n"
            f"{self.original_input}"
            )
        self.previous_transformation_result = str(self.ai.call(prompt))
    

    def call(self, input: str) -> str: # type: ignore
        """
        Perform a transformation on the input and return the result.

        Args:
            input (str): The input string.

        Returns:
            str: The output string.

        Raises:
            MaxTransformationAttemptsExceededError: If the maximum number of transformation attempts is exceeded.
        """
        self.original_input = input
        
        if self.attempts == self.max_attempts:
            raise MaxTransformationAttemptsMetError(attempts=self.attempts, max_attempts=self.max_attempts)
        else:
            self._come_up_with_transformation()
            self._perform_current_transformation()
            self.attempts += 1
            return self.previous_transformation_result
        
        
class MaxTransformationAttemptsMetError(Exception):
    """
    Exception raised when the maximum number of transformation attempts is exceeded.
    Attributes:
        attempts -- number of attempts made
        max_attempts -- maximum number of attempts allowed
    """
    def __init__(self, attempts: int, max_attempts: int):
        """
        Initialize the exception with the number of attempts made and the maximum number of attempts allowed.

        Args:
            attempts (int): The number of attempts made.
            max_attempts (int): The maximum number of attempts allowed.
        """
        self.message = f"Max transformation attempts of {max_attempts} exceeded. {attempts} attempts made."
        super().__init__(self.message)


from typing import Any, Callable, Type
from pydantic import BaseModel
from ai_alchemy.core.ai import AiWrapper, SmartAiWrapper, MaxTransformationAttemptsMetError
import logging
from dataclasses import dataclass
from copy import deepcopy

from typing import Protocol, TypeVar

class HasStrMethod(Protocol):
    """
    Protocol class for objects that have a __str__ method.
    """
    def __str__(self) -> str: ...

PydanticModel = TypeVar('PydanticModel', bound=BaseModel)

@dataclass
class TransformationContext:
    """
    Dataclass for storing the context of a transformation.

    Attributes:
    instruction: A string that describes the transformation task.
    output_type: The type of the output.
    output_schema: The schema of the output.
    input_data_type: The type of the input data.
    input_data: The input data.
    additional_context: Additional context for the transformation.
    final_instruction: The final instruction for the transformation.
    """
    instruction: str
    output_type: Type
    output_schema: str
    input_data_type: Type
    input_data: HasStrMethod
    additional_context: str
    final_instruction: str

    def __str__(self):
        """
        Returns a string representation of the TransformationContext.
        """
        base_prompt = [
            self.instruction,
            f"Here is the Output {self.output_type} schema:",
            self.output_schema,
            f"Here is the Input {self.input_data_type}:",
            str(self.input_data),
        ]

        if self.additional_context:
            base_prompt.append(f"Additional Context: {self.additional_context}")

        base_prompt.append(self.final_instruction)

        return "\n".join(base_prompt)
    
def _run_transformation(ai: AiWrapper, context: TransformationContext, validating_marshalling_function: Callable[[str], Any]) -> Any:
    """
    Runs a transformation using an AI model.

    Args:
    ai: The AI model to use for the transformation.
    context: The context of the transformation.
    validating_marshalling_function: A function that validates the output of the transformation.

    Returns:
    The result of the transformation.
    """
    

    prompt = str(context)
    ai_copy=deepcopy(ai)
    if type(ai_copy) == SmartAiWrapper:
        max_attempts = ai_copy.max_attempts
        num_attempts = 0
        while True:
            if num_attempts == max_attempts:
                raise MaxTransformationAttemptsMetError(attempts=num_attempts, max_attempts=max_attempts)
                # TODO This check logic is duplicated here from the smart ai wrapper, this is a better place to put it.
            try:
                import time
                time.sleep(1)

                initial = ai_copy.call(prompt)
                logging.debug(initial)
                if type(initial) == str:
                    r_val = validating_marshalling_function(initial)
                logging.debug(r_val)
                return r_val
            except Exception as e:

                num_attempts += 1
                ai_copy.previous_transformation_error = str(e)
                continue
    else:
        initial = ai_copy.call(prompt)
        logging.debug(initial)
        if type(initial) == str:
            r_val = validating_marshalling_function(initial)
        logging.debug(r_val)
        return r_val

def dict_to_pydantic_model(input: dict, ai: AiWrapper, output: Type[PydanticModel], additional_context:str="", strict: bool = True) -> PydanticModel:
    """
    Transforms a dictionary into a Pydantic model using an AI model.

    Args:
    input: The input dictionary.
    ai: The AI model to use for the transformation.
    output: The type of the Pydantic model to output.
    additional_context: Additional context for the transformation.
    strict: Whether to use strict validation for the Pydantic model.

    Returns:
    The transformed data as an instance of the ouput Pydantic Model.
    """
    context = TransformationContext(
        instruction="Given the following Pydantic schema and input dictionary, your task is to transform the dictionary into a JSON object that matches the Pydantic schema exactly. The transformed JSON object should not include the schema itself.",
        output_type=type(output),
        output_schema=str(output.model_json_schema()),
        input_data=input,
        input_data_type=type(input),
        additional_context=additional_context,
        final_instruction="Please provide the transformed JSON object."
    )

    def pydantic_validation_wrapper(data):
        return output.model_validate_json(json_data=data,strict=strict)
    
    return _run_transformation(ai=ai, context=context,validating_marshalling_function=pydantic_validation_wrapper)

def str_to_pydantic_model(input: str, ai: AiWrapper, output: Type[PydanticModel], additional_context:str="", strict: bool = True) -> PydanticModel:
    """
    Transforms a string into a Pydantic model using an AI model.

    Args:
    input: The input string.
    ai: The AI model to use for the transformation.
    output: The type of the Pydantic model to output.
    additional_context: Additional context for the transformation.
    strict: Whether to use strict validation for the Pydantic model.

    Returns:
    The transformed data as an instance of the ouput Pydantic Model.
    """
    context = TransformationContext(
        instruction="Given the following Pydantic schema and input string, your task is to transform the string into a JSON object that matches the Pydantic schema exactly. The transformed JSON object should not include the schema itself.",
        output_type=type(output),
        output_schema=str(output.model_json_schema()),
        input_data=input,
        input_data_type=type(input),
        additional_context=additional_context,
        final_instruction="Please provide the transformed JSON object."
    )

    def pydantic_validation_wrapper(data):
        return output.model_validate_json(json_data=data,strict=strict)
    
    return _run_transformation(ai=ai, context=context,validating_marshalling_function=pydantic_validation_wrapper)

def pydantic_model_to_pydantic_model(input: BaseModel, ai: AiWrapper, output: Type[PydanticModel], additional_context:str="", strict: bool = True) -> PydanticModel:
    """
    Transforms a Pydantic model into another Pydantic model using an AI model.

    Args:
    input: The input Pydantic model.
    ai: The AI model to use for the transformation.
    output: The type of the Pydantic model to output.
    additional_context: Additional context for the transformation.
    strict: Whether to use strict validation for the Pydantic model.

    Returns:
    The transformed data as an instance of the ouput Pydantic Model.
    """
    context = TransformationContext(
        instruction="Given the following Input data and Output Pydantic schema, your task is to transform the Input data into a JSON object that matches the Pydantic schema exactly. The transformed JSON object should not include the schema itself.",
        output_type=type(output),
        output_schema=str(output.model_json_schema()),
        input_data=input.model_dump_json(),
        input_data_type=type(input),
        additional_context=additional_context,
        final_instruction="Please provide the transformed JSON object."
    )
    def pydantic_validation_wrapper(data):
        return output.model_validate_json(json_data=data,strict=strict)
    
    return _run_transformation(ai=ai, context=context,validating_marshalling_function=pydantic_validation_wrapper)